/** @file ripper.cc */
//***************************************************************************
// This source code is copyrighted 2002 by Google Inc.  All rights
// reserved.  You are given a limited license to use this source code for
// purposes of participating in the Google programming contest.  If you
// choose to use or distribute the source code for any other purpose, you
// must either (1) first obtain written approval from Google, or (2)
// prominently display the foregoing copyright notice and the following
// warranty and liability disclaimer on each copy used or distributed.
// 
// The source code and repository (the "Software") is provided "AS IS",
// with no warranty, express or implied, including but not limited to the
// implied warranties of merchantability and fitness for a particular
// use.  In no event shall Google Inc. be liable for any damages, direct
// or indirect, even if advised of the possibility of such damages.
//***************************************************************************


// The Ripper iterates over one or more repository files, calling
// handler methods that act on each document in turn. Handlers are
// instances of a ParseHandler sub-class. See files
// parsehandler-caturl.cc and parsehandler-preparsecat.cc for examples
// of ParseHandler sub-classes.  To add a new handler to the ripper,
// look at the sections of this file labelled "*NEW HANDLERS*" and
// make the indicated modifications. Handlers make take arguments
// supplied on the ripper command line.
//
// Note: this is the public ripper, which operates on pre-parsed repositories.

#include <fstream>
#include <sstream>
#include <vector>
#include "handler-parser.h"
#include "repos-reader.h"
#include "parseelt.h"

// *NEW HANDLERS* - add includes here if needed
#include <fstream>
#include "graphbuilder.h"
#include "linkgraph.h"
#include "talker.h"

static void error(string errmsg) { cerr << errmsg << std::endl; exit(1); }

static void usage() {
  error("usage: ripper [--stop_after n] <handler options> "
        "{- | <repository files>}\n"
	"possible handler options are:\n"
	"--pvm_slave x         declares this ripper as a slave in a group of x tasks\n"
	"--pvm_master x        declares this ripper as the master in a group of x tasks\n"
	"--cat                 prints the full contents of each document\n"
	"--caturl              prints the url of each document\n"
	"--catlinks            prints the url and tolinks of each document\n"
	"--no_graphbuilder     doesn't build the web link graph\n"
	"--start_after x       starts numbering documents from x\n"
	"--stop_after x        stops processing after document number x\n"
	"--save_graph          prints the web link graph\n"
	"--save_index          prints an index of the documents read\n"
	"--string_memory x     reserves x Megabytes for all url strings\n"
	"--jumptable_memory x  reserves x Megabytes for all trie branch pointers\n"
	"--nodetable_memory x  reserves x Megabytes for the webnode hashtable\n"
	"--leaftable_memory x  reserves x Megabytes for the leafnode hashtable\n"
	"--name arthurdent     names this ripper arthurdent - no quotes please\n"
	"--temp_dir z          uses z as the temporary directory. Default is /tmp/\n"
	"--quit                noninteractive mode.\n"
);
}

static char* defaultrippername = "";
static char* defaulttempdir = "/tmp/";
int num_docs_processed = 0;

#define RIPPER_NAMELEN 20
#define RIPPER_TMPDIRLEN 20

/// Rips the repository and interfaces with main(). Calls GraphBuilder to build the web link graph
class Ripper {
 public:
  Ripper();
  ~Ripper();
  void SetupHandlers();
  void ParseCmdLineArgs(int argc, char** argv);
  void RipRepository (ReposReader* rr);
  void PrintStatistics(ostream& o);

  WebLinkGraph* PublishWebGraph() { return gb->UndockWebGraph(); }
  GraphBuilder* GetGraphBuilder() { return gb; }

  vector<string> rep_files_;

  // *NEW HANDLERS* - Add command line flags and values here
  struct {
    int stop_after;    ///< if non-zero, stop processing after this many docs
    int string_memory;    ///< memory to reserve for string table
    int jumptable_memory;    ///< memory to reserve for jumptable
    int nodetable_memory;    ///< memory to reserve for nodetable
    int leaftable_memory;   ///< memory to reserve for leaftable
    int start_ID;     ///< starting id number for WebNodes
    int pvm_numtasks; ///< number of pvm tasks in group
    bool pvm_is_master; ///< slave or master?
    bool repos_from_stdin; 
    bool interactive; ///< is ripper interactive or not?
    bool print_index; ///< prints the index of webnodes to a file
    bool handler_cat;  ///< simple handler to "cat" repository
    bool handler_caturl; ///< even simpler handler to "cat" just urls
    bool handler_catdate; ///< show date of the document
    bool handler_graph_print; ///< prints the web graph
    bool handler_catlinks; ///< display anchor links as they are being processed
    bool no_graphbuilder; ///< don't build the web link graph
  } flags_;

  char *rippername;
  char *tempdir;
  ofstream indexout;

 private:
  // list of parse-handlers to call for each document
  vector<ParseHandler*> parsehandlers_; 
  // this class builds the web graph from the repositories
  GraphBuilder* gb;
};

Ripper::Ripper() {
  flags_.stop_after = 0;
  flags_.start_ID = 0;

  flags_.string_memory  = 45; // Mb
  flags_.jumptable_memory  = 35; // Mb
  flags_.nodetable_memory  = 10; // Mb
  flags_.leaftable_memory  = 10; // Mb

  flags_.interactive = true;
  flags_.no_graphbuilder = false;
  flags_.repos_from_stdin = false;

  flags_.handler_cat = false;
  flags_.handler_caturl = false;
  flags_.handler_catdate = false;
  flags_.handler_graph_print = false;
  flags_.handler_catlinks = false;

  flags_.print_index = false;

  flags_.pvm_is_master = true;
  flags_.pvm_numtasks = 0;

  rippername = defaultrippername;
  tempdir = defaulttempdir;

  //  indexout = NULL;
  gb = NULL;
}

Ripper::~Ripper() {
  delete gb;
  // delete all parse handlers
  for (vector<ParseHandler*>::iterator ph = parsehandlers_.begin();
       ph != parsehandlers_.end(); ++ph)
    delete *ph;
}

void Ripper::SetupHandlers () {

  if(!flags_.no_graphbuilder) {
    gb = new GraphBuilder(flags_.string_memory, flags_.jumptable_memory,
			  flags_.nodetable_memory, flags_.leaftable_memory,
			  flags_.handler_catlinks);
  }

  if (flags_.handler_cat) {
    extern ParseHandler* MakeCatHandler();
    parsehandlers_.push_back(MakeCatHandler());
  }

  if (flags_.handler_caturl) {
    extern ParseHandler* MakeCatURLHandler();
    parsehandlers_.push_back(MakeCatURLHandler());
  }

  if (flags_.handler_catdate) {
    extern ParseHandler* MakeCatDateHandler();
    parsehandlers_.push_back(MakeCatDateHandler());
  }
  
  if (!flags_.no_graphbuilder) { // always make graph
    extern ParseHandler* MakeGraphHandler(GraphBuilder *graphbuilder);
    parsehandlers_.push_back(MakeGraphHandler(gb));
  }

  if (flags_.print_index) {
    if( !*rippername || (strlen(rippername) > RIPPER_NAMELEN) ) {
      cerr << "warning: index won't be saved. You must give ripper a short name." << endl;
      usage();
    } else {

      ostringstream nambuf;
      nambuf << tempdir << "ripper." << rippername << ".index";
      cerr << "info: index saved in file " << nambuf.str() << endl;
      indexout.open(nambuf.str().c_str());
      if( indexout.fail() ) {
 	cerr << "error: problem opening index file [" << nambuf.str() << "]" << endl;
      } 
    }
  }
  if (parsehandlers_.size() == 0) {
    usage();
  }

}

/// reads each document and builds a WebNode from
/// it through GraphBuilder
void Ripper::RipRepository (ReposReader* rr) {
  if( flags_.no_graphbuilder ) {

    while (!rr->AtEnd() && 
	   (flags_.stop_after == 0 || num_docs_processed < flags_.stop_after)) {
      ParseElt::Process_Document(rr, &parsehandlers_);
      num_docs_processed++;
    }

  } else {

    while (!rr->AtEnd() && 
	   (flags_.stop_after == 0 || num_docs_processed < flags_.stop_after)) {
      gb->NodeInitialize(num_docs_processed);
      ParseElt::Process_Document(rr, &parsehandlers_);
      gb->NodeInsertLinks();
      if( flags_.print_index && indexout) {
	 indexout << gb->NodeGetURL() << " " 
		  << gb->NodeGetID() << " "
		  << gb->NodeGetDate() << " " << gb->NodeGetAlias()
	   //		  << ((gb->NodeGetAlias()) ? gb->NodeGetAlias() : "") 
		  << " ** " << gb->NodeGetURL_() << " " << ((gb->NodeGetAlias_()) ? gb->NodeGetAlias_() : "")
<< endl;
      }
      gb->NodeLaunch();
      num_docs_processed++;
    }  

  }
}

/// prints out some memory usage statistics.
/// Useful for out of memory errors.
void Ripper::PrintStatistics(ostream& o) {
  o << "Number of documents processed:           " << num_docs_processed << endl;
  if(gb) 
    gb->StatisticsMem(o);
}

/// sets flags for all the possible command line options
void Ripper::ParseCmdLineArgs(int argc, char** argv) {
  argv++;
  for (int i = 1; i < argc; ++i, ++argv) {
    if ((*argv)[0] == '-' && (*argv)[1] != '\0') {  // option
      if (!strcmp(*argv, "--stop_after")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.stop_after = atoi(*argv);
      } else if (!strcmp(*argv, "--cat")) {
        flags_.handler_cat = true;
      } else if (!strcmp(*argv, "--caturl")) {
        flags_.handler_caturl = true;
      } 
      // *NEW HANDLERS* - add command line processing here
      else if (!strcmp(*argv, "--no_graphbuilder")) {
	flags_.no_graphbuilder = true;
      } else if (!strcmp(*argv, "--catdate")) {
	flags_.handler_catdate = true;
      } else if (!strcmp(*argv, "--quit")) {
	flags_.interactive = false;
      } else if (!strcmp(*argv, "--catlinks")) {
	flags_.handler_catlinks = true;
      } else if (!strcmp(*argv, "--save_graph")) {
	flags_.handler_graph_print = true;
      } else if (!strcmp(*argv, "--save_index")) {
	flags_.print_index = true;
      } else if (!strcmp(*argv, "--string_memory")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.string_memory = atoi(*argv);
      } else if (!strcmp(*argv, "--jumptable_memory")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.jumptable_memory = atoi(*argv);
      } else if (!strcmp(*argv, "--nodetable_memory")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.nodetable_memory = atoi(*argv);
      } else if (!strcmp(*argv, "--leaftable_memory")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.leaftable_memory = atoi(*argv);
      } else if (!strcmp(*argv, "--name")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        rippername = *argv;
	if( strlen(rippername) > RIPPER_NAMELEN) {
	  cerr << "Name too long, please use a shorter one." << endl;
	  usage();
	}
      } else if (!strcmp(*argv, "--temp_dir")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        tempdir = *argv;
	if( strlen(tempdir) > RIPPER_TMPDIRLEN) {
	  cerr << "Please use a shorter temporary directory name." << endl;
	  usage();
	}
      } else if (!strcmp(*argv, "--pvm_master")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.pvm_numtasks = atoi(*argv);
	flags_.pvm_is_master = true;
      } else if (!strcmp(*argv, "--pvm_slave")) {
        ++i;
        ++argv;
        if (i >= argc) {
          usage();
        }
        flags_.pvm_numtasks = atoi(*argv);
	flags_.pvm_is_master = false;
      } else if (!strcmp(*argv, "--start_after")) {
	++i;
	++argv;
	if (i >= argc) {
	  usage();
	}
	flags_.start_ID = atoi(*argv);
	num_docs_processed = flags_.start_ID;
      } else 
	usage();
    } else { // repository file
      if ((*argv)[0] == '-' && (*argv)[1] == '\0') {
        flags_.repos_from_stdin = true;
      } else {
        rep_files_.push_back(string(*argv));
      }
    }
  }
  if (!flags_.repos_from_stdin && rep_files_.empty()) {
    usage();
  }
  if (flags_.repos_from_stdin && !rep_files_.empty()) {
    cerr << "Specify only one source of repository input (files or stdin)"
         << std::endl;
    usage();
  }
}

Ripper *ripper = NULL;

/// Prints statistics when we run out of memory
void OutOfMemory() {
  cerr << "ran out of memory" << endl;
  if( ripper ) {
    ripper->PrintStatistics(cerr);
  }
  abort();
}


/// Reads the document repository and produces
/// a WebNode object for each read document. This
/// is handled by GraphBuilder. After the WebNodeList
/// (web link graph) is constructed, enters optionally
/// into interactive mode, by communicating with an
/// external program (jack) through pipes. 
int main(int argc, char** argv) {


  set_new_handler(OutOfMemory);
  ripper = new Ripper();

  cerr << "Welcome to the Google Programming Contest ripper." << std::endl
       << "Please see the file LICENSE for terms of use of "
       << "the data and code." << std::endl;

  ripper->ParseCmdLineArgs(argc, argv);
  ripper->SetupHandlers(); // opens index file

  // header for index file
  if (!ripper->flags_.no_graphbuilder &&
      ripper->flags_.print_index && ripper->indexout ) {
    ripper->indexout << "# Ripper: " << ripper->rippername << endl
		     << "# url | id | date | alias_url" << endl;
  }

  // this may take a long time to run
  if (ripper->flags_.repos_from_stdin) {
    ReposReader reprdr(&cin, string("<stdin>"));
    ripper->RipRepository(&reprdr);
  } else {
    for (vector<string>::iterator repname = ripper->rep_files_.begin();
         repname != ripper->rep_files_.end(); ++repname) {
      std::ifstream repstream((*repname).c_str());
      if (! repstream) {
        cerr << "error: Cannot open repository file " << *repname 
             << ", skipping it" << std::endl;
      } else {
        ReposReader reprdr(&repstream, *repname);
        ripper->RipRepository(&reprdr);
      }
      repstream.close();
    }
  }

  // close index file
  if (!ripper->flags_.no_graphbuilder &&
      ripper->flags_.print_index && ripper->indexout ) {
    ripper->indexout.close();
  }

  // interactive mode needs two fifos
  if(!ripper->flags_.no_graphbuilder && ripper->flags_.interactive) {

    cerr << "tempdir:" << ripper->tempdir <<endl;
    Talker *talker = new Talker(ripper->rippername, ripper->tempdir,
				ripper->PublishWebGraph(), 
				ripper->GetGraphBuilder(),
				ripper->flags_.pvm_is_master,
				ripper->flags_.pvm_numtasks);

    talker->Talk();
    delete talker;

  } else { // not interactive
    if( !ripper->flags_.no_graphbuilder &&
	ripper->flags_.handler_graph_print ) { 
      WebLinkGraph *graph = ripper->PublishWebGraph();
      graph->PrintWebGraph(cout);
    }
    ripper->PrintStatistics(cerr);
  }

  return 0;
}

