/** @file sampler.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef _SAMPLER_H_
#define _SAMPLER_H_
//#include "document.h"
#include "webnode.h"
#include "linkgraph.h"
#include <gsl/gsl_rng.h>


/// Abstract base class for all samplers. 
/**
 * A sampler
 * populates the web link graph with occupation counts
 * which are interpreted as page ranking schemes. 
 * Supported samplers must have transition probabilities
 * of the form P(x,y) = eps * mu(y) + (1-eps) * Q(x,y),
 * where the derived class overrides the Q(x,y) definition
 * which is contained in the QEvolveFrom() virtual function.
 */
class WebSampler {
 public:

  WebSampler(WebLinkGraph* g) throw (exception);


  void SimulateForward(long n);
  void ClearAllocForward();
  void SimulateAllocForward();
  void IncrementAllocForward(uint32 k, int32 num);

  void TaggedSimulateForward(long n, const uint32 *fromsetsize);
  /// performs a random transition, should be overridden by derived classes
  virtual WebNodePtr QEvolveFrom(WebNodePtr x) 
    { return x; } 
  /// returns an identifying string in supplied buffer
  virtual char *Name(char *buf) = 0;
  /// Clears occupation counts and run sizes
  void ClearCounts();
  void ClearScratch();
  void SetRandomSeed(unsigned long int s) { gsl_rng_set(r,s); }
  void PrintCounts(ostream& o);
  void PrintTagCounts(ostream& o);
  /// Number of samples produced by SimulateForward()
  uint32 LastRunSize() 
    { return last_run_size; }
  /// Number of candidates accepted by TaggedSimulateForward()
  uint32 LastTaggedRunSize() 
    { return last_tagged_run_size; }

 protected:
  WebLinkGraph* graph;
  uint32 number_of_nodes;

  gsl_rng* r; // for random numbers
  double eps;

  int32 *allocation_list; // for SimulateForward()
  LeafNodePtr xleaf;

  double probabilities[TAG_NUMBER_OF_BITS]; // for TaggedSimulateForward()
  long allocated[TAG_NUMBER_OF_BITS];
  uint32 fromsetsize_remaining[TAG_NUMBER_OF_BITS];

  uint32 last_run_size;
  uint32 last_tagged_run_size;
};


/// A WebSampler which implements the standard PageRank 
/// chain.
class PageRankSampler: public WebSampler {
 public:
  PageRankSampler(WebLinkGraph* agraph);
  void SetParameters(double epsilon);
  WebNodePtr QEvolveFrom(WebNodePtr x);
  virtual char *Name(char *buf);

};

/// A scratch structure used by DateBiasedPageRankSampler
typedef union DBScratchStruct {

  DBScratchStruct(ScratchStruct s) 
  { ss = s; }
  ~DBScratchStruct() {}

  struct {
    uint16 filled;
    int16 minvalue;
  } db;

  float mass;
    
  ScratchStruct ss;
};

/// A WebSampler which implements page ranking by taking 
/// into account the relative age of linked-to documents.
class DateBiasedPageRankSampler: public WebSampler {
 public:
  DateBiasedPageRankSampler(WebLinkGraph* agraph);
  void SetParameters(double epsilon, double lambda, uint16 daterange);
  WebNodePtr QEvolveFrom(WebNodePtr x);
  virtual char *Name(char *buf);
 private:
  double lam;
  float lamhat;
  uint16 daterange;
};

typedef enum { undef, hubs, auth } ktype;

/// A WebSampler which implements an epsilon approximation 
/// of the Kleinberg hubs and authorities model.
/**
 * The Kleinberg hubs and authorities model can be obtained by 
 * running a Markov chain with transition probs QQ^*, where Q 
 * follows tolinks at random, and Q^* follows fromlinks at random.
 * This chain is not UE, so doesn't fit into our framework. But
 * by choosing epsilon close to zero, we can get an approximation.
 *
 */
class TruncatedKleinbergSampler: public WebSampler {
 public:
  TruncatedKleinbergSampler(WebLinkGraph* agraph);
  void SetParameters(double epsilon, ktype what);
  WebNodePtr QEvolveFrom(WebNodePtr x);
  virtual char *Name(char *buf);
 private:
  ktype which;
};

#endif
