/** @file webnode.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer (laird@lbreyer.com)
 */

#include "webnode.h"

// WebNode objects, which are of type MemoryPooled<WebNodeStruct>,
// are allocated by their own memory manager, part of which is declared here
const int MemoryPooled<WebNodeStruct>::BLOCK_SIZE=WEBNODE_MEMPOOL_DELTA;
MemoryPooled<WebNodeStruct> *MemoryPooled<WebNodeStruct>::freelist = NULL;

// Link objects, which are of type MemPool<LinkStruct>,
// are allocated by their own memory manager
MemPool<LinkStruct> WebNode::global_link_pool;

// the global_link_pool contains a few static data members
const int MemPool<LinkStruct>::BLOCK_SIZE = LINK_MEMPOOL_DELTA;
MemPoolObject<LinkStruct>* MemPool<LinkStruct>::mempool1 = NULL;
MemPoolObject<LinkStruct>* MemPool<LinkStruct>::mempool2 = NULL;


WebNode :: WebNode(uint32 idno) {
  data.id = idno;
  data.date = 0;
  data.num_valid_tolinks = 0;
  data.num_leaflinks = 0;
  data.num_tolinks = 0;
  data.num_fromlinks = 0;
  data.tolinks = NULL;
  data.fromlinks = NULL;
}

/// Sets the earliest known date of the WebNode
/**
 * This function is designed to be called several times.
 * The earliest nonzero date is retained.
 */
void WebNode :: SetDate(uint16 adate) {
  if( adate > 0 ) {
    if( data.date > 0 ) {
      data.date = min(data.date,adate);
    } else {
      data.date = adate;
    }
  }
}

/// Inserts a set of tolinks into the WebNode.
/**
 * Inserts (or merges) the raw (character ptr) links
 * into the webnode's tolinks array.
 */
void WebNode :: InsertRawLinks(RawLinkSet *s) {
  assert(data.num_valid_tolinks == 0);
  assert(data.num_leaflinks == 0);
  if(data.num_tolinks > 0) {
    for(int j = 0; j < data.num_tolinks; j++) {
      s->insert(data.tolinks[j].data.pointer_diff);
    }
    WebNode::global_link_pool.Deallocate(data.tolinks,data.num_tolinks);
    data.num_tolinks = 0;
  }

  data.num_tolinks = s->size();
  if( data.num_tolinks > 0 ) {
    data.tolinks = WebNode::global_link_pool.Allocate(data.num_tolinks);
    int k = 0;
    for(RawLinkSet::iterator i = s->begin(); i != s->end(); i++) {
      data.tolinks[k++].data.pointer_diff = (*i);
    }
    assert(k == data.num_tolinks);

  } else {

    data.tolinks = NULL;

  }
}


/// This appends a webnode to the fromlinks list.
void WebNode :: AppendFromLink(WebNodePtr anothernode) throw (overflow_error) {

  assert(anothernode);
  if( data.num_fromlinks <= 0 ) { 
    cerr << "error: trying to append to empty fromlink" << endl;
    throw overflow_error("");
  }

  if( data.fromlinks == NULL ) {
    data.fromlinks = WebNode::global_link_pool.Allocate(data.num_fromlinks);
    assert(data.fromlinks != NULL);
    for(int k = 0; k < data.num_fromlinks; k++) {
      data.fromlinks[k].data.webnode_ptr = NULL;
    }
  }


  // insert anothernode in the first free nonempty slot. 
  for(int k = 0; k < data.num_fromlinks; k++) {
    if(data.fromlinks[k].data.webnode_ptr == NULL) {
      data.fromlinks[k].data.webnode_ptr = anothernode;
      return;
    }
  }
  // if we get to here, there's a problem
  cerr << "error: fromlink overflow" << endl;
  throw overflow_error("");
}

/// Sorts all valid links into the first part of the tolinks array.
/**
 * While sorting, it also
 *  converts the pointer_differences into webnode_ptrs.
 * Dangling links are left in the upper half of the array.
 *
 * Note second argument should really be a data member of first
 * (more elegant) but then we'd have to create a derived simplehashtable...
 */
void WebNode :: NormalizeRawLinks(SimpleHashTable<WebNodePtr> *h) {
  int i, j;
  for(i = 0; i < data.num_tolinks; i++) {
    WebNodePtr w = h->Find(data.tolinks[i].data.pointer_diff);
    if( w ) {
      data.tolinks[i].data.webnode_ptr = w;
    } else { // its dangling link
     // see if another valid link exists
      for(j = i + 1; j < data.num_tolinks; j++) {
	WebNodePtr v = h->Find(data.tolinks[j].data.pointer_diff);
	if( v ) { // found one, swap
	  data.tolinks[j].data.pointer_diff = data.tolinks[i].data.pointer_diff;
	  data.tolinks[i].data.webnode_ptr = v;
	  break;
	}
      }
      if( j == data.num_tolinks ) { // all other links were dangling
	// so all elements i, i+1, i+2,... are dangling links
	break;
      }
    }
  }

  data.num_valid_tolinks = i;

#ifndef NDEBUG
  // consistency check/validation
  for(int k = 0; k < NumberOfValidToLinks(); k++) {
    assert(ValidToLink(k));
    assert(ValidToLink(k)->OccupationCount() == 0);
  }
  for(int k = NumberOfValidToLinks(); k < data.num_tolinks; k++) {
    assert(!h->Find(data.tolinks[k].data.pointer_diff));
  }
#endif
}

/// Returns the full size of the WebNode including link arrays.
size_t WebNode :: RealSize() {
  return sizeof(WebNode) + sizeof(Link) * data.num_tolinks 
    + sizeof(Link) * data.num_fromlinks;
}


/// sifts LeafNode pointers upward in the tolinks array.
/**
 * The code for this function is very similar to that of NormalizeRawLinks()
 */
void WebNode :: UpdateLeafLinks(SimpleLeafNodePtrHashTable *leaftable) {
  int i, j;
  for(i = data.num_valid_tolinks + data.num_leaflinks; i < data.num_tolinks; i++) {
    LeafNodePtr w = leaftable->Find(data.tolinks[i].data.pointer_diff);
    if( w ) {
      data.tolinks[i].data.leafnode_ptr = w;
    } else { // its dangling link
     // see if another leafnode link exists
      for(j = i + 1; j < data.num_tolinks; j++) {
	LeafNodePtr v = leaftable->Find(data.tolinks[j].data.pointer_diff);
	if( v ) { // found one, swap
	  data.tolinks[j].data.pointer_diff = data.tolinks[i].data.pointer_diff;
	  data.tolinks[i].data.leafnode_ptr = v;
	  break;
	}
      }
      if( j == data.num_tolinks ) { // all other links were dangling
	// so all elements i, i+1, i+2,... are dangling links
	break;
      }
    }
  }

  data.num_leaflinks = i - data.num_valid_tolinks;

#ifndef NDEBUG
  // consistency check/validation
  for(int k = 0; k < NumberOfLeafLinks(); k++) {
    assert(ValidLeafLink(k));
    assert(ValidLeafLink(k)->OccupationCount() == 0);
  }
  for(int k = NumberOfValidToLinks() + NumberOfLeafLinks(); k < data.num_tolinks; k++) {
    assert(!leaftable->Find(data.tolinks[k].data.pointer_diff));
  }
#endif
  
}
